"""Module defining the caching strategy for overpass requests."""
import os
import xml.etree.ElementTree as ET
import hashlib

from ..constants import OSM_CACHE_DIR


def get_cache_key(query: str) -> str:
    """
    Generate a unique cache key for the query using a hash function.
    This ensures that queries with different parameters are cached separately.
    """
    return hashlib.md5(query.encode('utf-8')).hexdigest()


class CachingStrategyBase:
    """
    Base class for implementing caching strategies.

    This class defines the structure for a caching strategy with basic methods
    that must be implemented by subclasses. Subclasses should define how to
    retrieve, store, and close the cache.
    """
    def get(self, key):
        """Retrieve the cached data associated with the provided key."""
        raise NotImplementedError('Subclass should implement get')

    def set(self, key, value):
        """Store data in the cache with the specified key."""
        raise NotImplementedError('Subclass should implement set')

    def close(self):
        """Clean up or close any resources used by the caching strategy."""


class XMLCache(CachingStrategyBase):
    """
    A caching strategy that stores and retrieves data in XML format.

    This class provides methods to cache data as XML files in a specified directory.
    The directory is automatically suffixed with '_XML' to distinguish it from other
    caching strategies. The data is stored and retrieved using XML serialization.

    Args:
        cache_dir (str): The base directory where XML cache files will be stored.
                         Defaults to 'OSM_CACHE_DIR' with a '_XML' suffix.

    Methods:
        get(key): Retrieve cached data from a XML file associated with the given key.
        set(key, value): Store data in a XML file with the specified key.
    """
    def __init__(self, cache_dir=OSM_CACHE_DIR):
        # Add the class name as a suffix to the directory
        self._cache_dir = f'{cache_dir}_XML'
        if not os.path.exists(self._cache_dir):
            os.makedirs(self._cache_dir)

    def _filename(self, key):
        return os.path.join(self._cache_dir, f'{key}.xml')

    def get(self, key):
        """Retrieve XML data from the cache and parse it as an ElementTree."""
        filename = self._filename(key)
        if os.path.exists(filename):
            try:
                # Parse and return the cached XML data
                tree = ET.parse(filename)
                return tree.getroot()  # Return the root element of the parsed XML
            except ET.ParseError:
                # print(f"Error parsing cached XML file: {filename}")
                return None
        return None

    def set(self, key, value):
        """Save the XML data as an ElementTree to the cache."""
        filename = self._filename(key)
        tree = ET.ElementTree(value)  # value is expected to be an ElementTree root element
        try:
            # Write the XML data to a file
            with open(filename, 'wb') as file:
                tree.write(file, encoding='utf-8', xml_declaration=True)
        except IOError as e:
            raise IOError(f"Error writing to cache file: {filename} - {e}") from e


class CachingStrategy:
    """
    A class to manage different caching strategies.

    This class provides an interface to switch between different caching strategies 
    (e.g., XMLCache, JSONCache) dynamically. It allows caching data in different formats, 
    depending on the strategy being used. By default, it uses the XMLCache strategy.

    Attributes:
    __strategy (CachingStrategyBase): The currently active caching strategy.
    __strategies (dict): A mapping between strategy names (as strings) and their corresponding
                         classes, allowing dynamic selection of caching strategies.
    """
    __strategy = XMLCache()  # Default caching strategy
    __strategies = {
        'XML': XMLCache,
    }

    @classmethod
    def use(cls, strategy_name='XML', **kwargs):
        """
        Set the caching strategy based on the strategy_name provided.

        Args:
            strategy_name (str): The name of the caching strategy (e.g., 'XML').
            **kwargs: Additional keyword arguments to pass when initializing the strategy.
        """
        # If a previous strategy exists, close it
        if cls.__strategy:
            cls.__strategy.close()

        # Retrieve the strategy class based on the strategy name
        strategy_class = cls.__strategies.get(strategy_name)

        if not strategy_class:
            raise ValueError(f"Unknown caching strategy: {strategy_name}")

        # Instantiate the new strategy with the provided arguments
        cls.__strategy = strategy_class(**kwargs)
        return cls.__strategy

    @classmethod
    def get(cls, key):
        """Get data from the current strategy's cache."""
        if not cls.__strategy:
            raise RuntimeError("Caching strategy has not been set.")
        return cls.__strategy.get(key)

    @classmethod
    def set(cls, key, value):
        """Set data in the current strategy's cache."""
        if not cls.__strategy:
            raise RuntimeError("Caching strategy has not been set.")
        cls.__strategy.set(key, value)
